// this is your game's Cellscript, used to instruct the Cell compiler how to
// build your game.  like a Sphere game, the Cellscript is written in
// JavaScript, however with a different set of bindings tailored for compiling
// and packaging games.


// describe the game we're building.  this gets written to the game manifest
// (game.json) at the end of the build.
import { convertRSS } from "./src/rssLoader.mjs";
import { convertRMP } from "./src/rmpLoader.mjs";


Object.assign(Sphere.Game, {
	name: "T",
	version: 2,
	author: "???",
	summary: "",
	resolution: '512x384',
	main: 'src/main.mjs',
});

let rssTool = new Tool(convertRSS, "converting RSS");
let rmpTool = new Tool(convertRMP, "converting RMP");

//simple function to run the above tools on collections of files
function runTool(dirName, sources, tool, extension)
{
    const targets = [];
    FS.createDirectory(dirName);
    for (const source of sources)
    {
        let fileName = FS.fullPath(source.name, dirName);
        fileName = fileName.substring(0, fileName.lastIndexOf(".")) + extension;
        targets.push(tool.stage(fileName, [source], { name : source.name }));
    }
    return targets;
}

//update paths in the second parameter to the folders you're using
runTool("@/sprites", files("sprites/*.rss"), rssTool, ".ses");
runTool("@/maps", files("maps/*.rmp"), rmpTool, ".mem");

//update path in second parameter to folder your mapscripts are stored in
install("@/maps/scripts", files("src/mapScripts/*.mjs"));
install("@/bin", files("src/*.mjs"));
install("@/shaders",  files("shaders/*.glsl", true));

//remove this line if not using windowstyles
install("@/windows",  files("windows/*.rws", true));


install('@/fonts/', files('fonts/*.rfn', true));
install('@/images/', files('images/*.png', true));
install('@/maps/', files('maps/*.rmp', true));
install('@/maps/', files('maps/*.rts', true));
install('@/src/', files('src/*.js', true));
install('@/src/', files('src/*.mjs', true));
install('@/sounds/', files('sounds/*.wav', true));
install('@/spritesets/', files('spritesets/*.rss', true));
